from .abstract_field import AbstractField
from pyparsing import ParseException
class Field(AbstractField):
    def __init__(self,s,lok,toks):
        super().__init__(level=int(toks["level"]),name=toks["name"])
        self.toks = toks
        self.list_of_values = {}
        self.redefine_target = toks.get("redefine_target")
        self.repeat = int(toks.get("occurs","1"))
        self.length = 0
        self.precision = 0
        self.signed = False
        self.compressed = False
        self.signed_trailing = False
        self.explicit_decimal = False

        if toks.get("type_numeric") is not None:
            type_descriptor = toks.get("type_numeric")
            self.length = int(type_descriptor.get("length","0"))
            # handle precision
            self.precision = int(type_descriptor.get("precision","0"))
            if self.precision>0:
                self.datatype = "decimal"
                if type_descriptor.get("explicit_decimal"):
                    self.explicit_decimal = True
            else:
                self.datatype = "int"
            
            # signed
            if type_descriptor.get("signed"):
                self.signed = True
                if toks.get("trailing"):
                    self.signed_trailing = True

            # compressed
            if type_descriptor.get("comp3"):
                self.compressed = True

            # validate
            if self.length==0 and self.precision==0:
                raise ParseException(f"can't parse PIC for {toks}")
        elif toks.get("type_string"):
            # a string PIC
            type_descriptor = toks.get("type_string")
            self.length = int(type_descriptor.get("length","0"))

            self.datatype = "str"

            # check if we have a list of possible values. convert that to a nice constants array
            if toks.get("values"):
                for val in toks.get("values"):
                    self.list_of_values[val.get("name")] = val.get("value")
        else:
            # no type descriptor
            raise ParseException(f"unknown data type: {toks}")
    def get_total_length(self):
        return (1 if self.signed else 0) +\
            (1 if self.explicit_decimal else 0) +\
            self.length +\
            self.precision 

    def parse(self,input:str):
        if self.datatype=="str":
            return input
        elif self.datatype=="int":
            return int(input)
        elif self.datatype=="decimal":
            # put a decimal notation and parse as float
            return float(f"{input[0:self.length]}.{input[-self.precision:]}")

    def __repr__(self):
        return_value = f"name: {self.name}\n- type: {self.datatype}\n- length: {self.length}\n- precision {self.precision}"
        if self.list_of_values != {}:
            for k,v in self.list_of_values.items():
                return_value += f"\n  - {k}:{v}"
        return return_value



# class representing a field or field group
class AbstractField:
    def __init__(self,level:int,name:str):
        self.level = level
        self.name = name
        self.start_pos:int = 0
        self.length:int = 0

    def __repr__(self):
        return  f"name: {self.name}"

    def _rename_with_suffix(self,suffix:str):
        # in cases of reoccurs we want to normalize the field names so they are unique per record. useful for parsing
        self.name += suffix

    def get_total_length(self) -> int:
        return 0



from pyparsing import *
from .field_group import FieldGroup
from .field import Field

# main pyparsing grammar definitions for the copybook format
stmt = Forward()
LPAR, RPAR, COMMA = map(Suppress, "(),")

char_pic = Group(
    # bracket notation, e.g. X(2)
    (Suppress("X")
    + LPAR
    + Word(nums)("length")
    + RPAR)
    # picture notation, e.g. XX
    | OneOrMore("X").setParseAction(lambda tokens: len(tokens))("length")
)("type_string")

numeric_pic = Group(
    Optional(
        Suppress("S")("signed")
        | (one_of("+ -")+FollowedBy("9"))("signed")
        | (one_of("+ -")+FollowedBy(".9"))("signed")
        | (one_of("+ -")+FollowedBy("V9"))("signed")
    )
    + Optional(
        # bracket notation, e.g. 9(13)
        (
            Suppress("9")
            + LPAR
            + Word(nums)("length")
            + RPAR
        )
        # picture notation, e.g. 9999
        | OneOrMore("9").setParseAction(lambda tokens: len(tokens))("length")
    )
    + Optional(
        (
            ("V"+FollowedBy("9"))("implied_decimal")
            | ("."+FollowedBy("9"))("explicit_decimal")
        )
        # bracket notation, e.g. 9(13)
        + (
            (
            Suppress("9")
            + LPAR
            + Word(nums)("precision")
            + RPAR
            )
            # picture notation, e.g. 9999
            | OneOrMore("9").setParseAction(lambda tokens: len(tokens))("precision")
        )
    )
    + Optional(
        Literal("SIGN")("signed")
        + (
            Literal("TRAILING")("trailing") | Literal("LEADING")("leading")
        )
        + Literal("SEPARATE")
    )
    + Optional(
        Suppress("COMP-3")("comp3")
    )
)

pic_expr = (
    "PIC" +
    (char_pic | numeric_pic("type_numeric"))
)
field_title = Word(alphanums+"-_")

field_group = (
    Word(nums)("level") 
    + field_title("name") 
    + Optional(
        Suppress("REDEFINES")
        + Word(alphanums+"-")("redefine_target")
    )
    + Optional(
        "OCCURS"
        + Word(nums)("occurs")
        + Optional("TIMES")
    )
    + "."
)

# the 88 field type is an indicator of possible list of values
QuotedString("'")("value")

field = (
    Word(nums)("level") 
    + field_title("name") 
    + Optional(
        Suppress("REDEFINES")
        + Word(alphanums+"-")("redefine_target")
    )
    + Optional(
        "OCCURS"
        + Word(nums)("occurs")
        + Optional("TIMES")
    )
    + pic_expr
    + Optional(
        "VALUE"
        + SkipTo(".")
    )
    + "."
    + ZeroOrMore(
        Group(Suppress("88")
        + Word(alphanums+"-")("name") 
        + Suppress("VALUE")
        + SkipTo(".")("value")
        + "."
    )("constant"))("values")
)

non_field_row = Group(
    Word(alphas) +
    ... + 
    "."
)
EOF_CHR = "\x1a"
stmt << (
    Suppress(
        SkipTo( Combine(LineStart() +" "*6 + Char(" *") + ZeroOrMore(White()) + Keyword("01") ).leaveWhitespace())
    )
    + OneOrMore(
        (
            field_group.setParseAction(FieldGroup)
            | field.setParseAction(Field)
        )
    )
    + Suppress(ZeroOrMore(
        non_field_row
    ))
    + ZeroOrMore(EOF_CHR)

)



# class representing a field group
import copy
from collections import namedtuple
from .abstract_field import AbstractField
class FieldGroup(AbstractField):
    def __init__(self,s,lok,toks):
        super().__init__(level=int(toks["level"]),name=toks["name"])
        self.redefine_target = toks.get("redefine_target")
        self.repeat = int(toks.get("occurs","1"))
        self.children = []

    def calculate_positions(self,start_pos=0):
        # take care of any reoccurs lurking out there
        self._normalize_reoccurs()
        self.start_pos = start_pos
        # loop through the children and update the positions
        for child in self.children:
            # handle redefines
            if child.redefine_target is not None:
                # redefine means the position is that of a group we already encountered. find it
                target = self.get_child_by_name(child.redefine_target)
                if type(child)==FieldGroup:
                    child.calculate_positions(target.start_pos)
                else:
                    child.start_pos = target.start_pos
                # skip to the next without advancing the start_pos. do not pass go, do not collect $200
                continue

            if type(child)==FieldGroup:
                child.calculate_positions(start_pos)
                # handle redefines and occurs
            child.start_pos = start_pos
            start_pos += child.get_total_length()

    def _rename_with_suffix(self,suffix:str):
        # in cases of reoccurs we want to normalize the field names so they are unique per record. useful for parsing
        super()._rename_with_suffix(suffix)
        for child in self.children:
            child._rename_with_suffix(suffix)

    def _normalize_reoccurs(self):
        # this method duplicates the reoccurs and renames the fields sequentially to avoid duplicate names
        childcount:int = len(self.children)
        for i in range(childcount,0,-1):
            # we reverse loop since we are changing the list within the loop
            child = self.children[i-1]
            if child.repeat>1:
                # deep copy
                for clone_number in range(child.repeat,1,-1):
                    newchild = copy.deepcopy(child)
                    newchild._rename_with_suffix("_"+str(clone_number))
                    # push into the list
                    self.children.insert(i,newchild)


    def flatten(self):
        # flatten the tree into a flat list of Fields
        return_list = [self]
        for child in self.children:
            if type(child)==FieldGroup:
                return_list.extend(child.flatten())
            else:
                return_list.append(child)
        return return_list

    def get_child_by_name(self,name:str):
        return next(filter(lambda x: x.name==name,self.children),None)

    def get_total_length(self) -> int:
        total_length = 0
        for child in self.children:
            # if this is a 'redefine' it does not occupy any space
            if child.redefine_target is None:
                total_length += child.get_total_length()

        return total_length

    def __repr__(self):
        return_value = f"""
name: {self.name}
- type: group
- redefines: {self.redefine_target}
"""
        return return_value




import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="copybook",
    version="1.0.15",
    description="python copybook parser",
    long_description=README,
    
    long_description_content_type="text/markdown",
    url="https://github.com/zalmane/copybook",
    author="Oren Elias",
    author_email="support@datayoga.io",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["copybook"],
    include_package_data=True,
    install_requires=['pyparsing'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
)